// |$Id: mseed2seg.c,v 1.1.1.1 2022/12/27 22:06:24 pm Exp $         |
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              |
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS    http://173.255.241.228              | 
   | FUNCTION NAME: mseed2seg.c                                 |
   | DIRECTORY: bsu-3.0.3/src/C/bsegy                           |
   |------------------------------------------------------------|
   | purpose: convert MSEED format data (Iris typical)          |
   |          to BSEGY formatted data (limited by number of     |
   |          samples in a BSEGY header                         |
   |------------------------------------------------------------|
   |  NOTICE:                                                   |
   |  Copyright (C) 2021 Paul Michaels                          |
   |  <paulmichaels@boisestate.edu>                             |
   |  This program is free software; you can                    |
   |  redistribute it and/or modify it under the terms          |
   |  of the GNU General Public License as published            |
   |  by the Free Software Foundation; either version           |
   |  3 of the License, or (at your option) any later           |
   |  version.  This program is distributed in the              |
   |  hope that it will be useful, but                          |
   |  WITHOUT ANY WARRANTY; without even the implied            |
   |  warranty of MERCHANTABILITY or FITNESS FOR A              |
   |  PARTICULAR PURPOSE.  See the GNU General Public           |
   |  License for more details.                                 |
   |  You should have received a copy of the GNU                |
   |  General Public License along with this program;           |
   |  if not, write to the Free Software Foundation,            |
   |  Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             |
   |------------------------------------------------------------|
   |INLINE FUNCTIONS:                                           |
   |  getparm    get parameters from command line or interactive|
   |  outlst     echo check of input parameters to listing file |
   |                                                            |
   |------------------------------------------------------------|
   |sublibC4 functions:                                         |
   |  c_bsegin   read seismic trace                             |
   |  c_bsegout  write a seismic trace                          |
   |  in_chk     scan input file for samples, number traces etc.|
   |  bargrid    write progress bar grid to screen              |
   |  exbar      write progress bar to screen                   |
   |  cr_labl    write copyright label to screen                |
   |  fcr_labl   write copyright label to file                  |
   |  names      create an output file name from input file name|
   |                                                            |
   |------------------------------------------------------------|
*/
      #include<errno.h>
      #include <stdio.h>
      #include <stdlib.h>
      #include <float.h>
      #include <math.h>
      #include <string.h>
      #include <time.h>
      #include "../include/c_bsegy.h"

	#ifndef WIN  
	#include <signal.h>
	static void term_handler (int sig);
	#endif

	#include <libmseed.h>


//...parameter definitions

//...function prototypes
      #include "../include/sub4.h"

      int  getparm(int argc,char *argv[],
      char infile[], char wantStation[], char wantchannel[],char wantnetwork[],
      char wantlocation[], double *wantsamprate, int *scansw, 
      char ofile[], char lstfile[],char datfile[],char pid[]);

      void names2(char filename[],char wantStation[],
      char wantchannel[], char sfx[]);

      void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], int scansw,
	char wantStation[],char wantchannel[], char wantnetwork[], char wantlocation[],
	double wantsamprate,char datfile[], FILE *h3);

	int scanning (char infile[], MSRecord *msr, int *reccount, int scansw);

	int readingdata (char infile[], MSRecord *msr, int *reccount, int scansw,
	char wantStation[], char wantchannel[], char wantnetwork[],char wantlocation[],
	double wantsamprate, float fsamin, char datfile[], int *npts, FILE *h1);

	void  genhead(int jrec, float fsamin, int npts, float samprate, char wantchannel[],
	char wantStation[], 
        uint16_t year, uint16_t day, uint8_t hour, uint8_t min, uint8_t sec,
        struct header *hd);	

	int readingvalues (char infile[], MSRecord *msr, int *reccount, int scansw, int npts,
        uint16_t *year, uint16_t *day, uint8_t *hour, uint8_t *min, uint8_t *sec,
	char wantStation[], char wantchannel[], char wantnetwork[], char wantlocation[],
	double wantsamprate, float fsamin, float *s2 );

	short doy2moday(int doy,  int year);

	void	gengnuplot(char plotfile[], char wantStation[], char wantchannel[],
	uint16_t year, uint16_t day, uint8_t hour, uint8_t min, uint8_t sec,
	char datfile[]);

	int	findnptshort(int npts);



static flag verbose = 0;
static flag ppackets = 0;
static flag basicsum = 0;
static int printdata = 0;  // 1 if printing data out
static int reclen = -1;
static char *inputfile = 0;
static int ndim = 65536;  //maximum number of samples in a BSEGY file

static void term_handler (int sig);

      int  main(int argc, char *argv[])
     {

//...declare variables
      FILE              *h1, *h2, *h3;
      int               jt,
                        jrec,
                        npts,
                        ntraces,
                        ibar,
                        invbar;
      float             fsamin,
                        bar,
                        *s1,
                        *s2,
			*p0;  //pointer to first sample in S2
      char              pid[5]="msed",   //process id name (4char)
			plotfile[13],
                        ofile[13],
			datfile[13],
                        lstfile[13],
                        infile[132];  //OK, MSEED files likely to have a long name
//                        infile[14]; //making an exception.  Outfile will take only
      struct header     hd;           //first 4 characters of the long name for output names.
				      //why 132?  Back in the 1960's that was a common size.	
//...mseed variables
	MSRecord *msr = NULL;
	int64_t totalrecs = 0;
	int64_t totalsamps = 0;
	int retcode;
	int itest0 = 1;
	int itest1 = 1;
	int itest2 = 1;
	int itest3 = 1;
	hptime_t START;
	int reccount = 0;

//...specific to this application
	char wantStation[11];
	char wantchannel[11];
	char wantlocation[11];
	char wantnetwork[11];
	double wantsamprate;
	float samprate;
	int scansw;
	double T;
	uint16_t year;
	uint16_t day;
	uint8_t hour;
	uint8_t min;
	uint8_t sec;
	int	nptshort;
	ntraces = 1;


//...initialize number of samples counter
	npts = 0;

//...get parameters
      if(getparm(argc,argv,infile,wantStation,wantchannel,wantnetwork,
	wantlocation,&wantsamprate,&scansw,
        ofile,lstfile,datfile,pid)!=0)
      {
       goto quick_exit;
      }

//...set sample interval
	fsamin = 1.0/wantsamprate;

//...open output listing file
        if( (h3=fopen(lstfile,"w"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",lstfile);
          goto quick_exit;
        }

//...echo check to listing file
        outlst(infile,pid,ofile,lstfile,scansw,wantStation,
	wantchannel,wantnetwork,wantlocation,wantsamprate,datfile,h3);
	fflush(h3);

//...redirect stdout to listing file for use of msr_print()
	freopen(lstfile,"a+",stdout);

//...Case Scan Only
	reccount = 0;
	if (scansw != 0 ) {
	  retcode = scanning(infile,msr,&reccount,scansw); 
}
else {
//...open output data file
        if( (h1=fopen(datfile,"w"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",datfile);
          goto quick_exit;
        }
	  retcode = readingdata(infile,msr,&reccount,scansw,
	  wantStation,wantchannel,wantnetwork,wantlocation,wantsamprate,
	  fsamin,datfile,&npts,h1);

} // end scan only switch else block

#ifndef WIN
	freopen("/dev/tty","w",stdout);	//linux
#else
	freopen("CON","w", stdout); //Mingw C++, Windows
#endif

  if ( retcode != MS_ENDOFFILE )
    ms_log (2, "Cannot read %s: %s\n", infile, ms_errorstr(retcode));

  /* Cleanup memory and close file */
  ms_readmsr (&msr, NULL, 0, NULL, NULL, 0, 0, verbose);
  
//...Check maximum number of samples extracted with BSEGY limit
	
if (scansw == 0 ) {
	if (npts <= 65535) {
	fprintf(stdout,"|   Good, number of samples = %d which is less than 65535\n",npts);
	fprintf(stdout,"|   Output will be 1 trace\n");
//  	fprintf(h3,"==> Good, number of samples = %d which is less than 65535\n",npts);
//  	fprintf(h3,"==> Output will be 1 trace\n");
	}
	else {
	fprintf(stdout,"|   Large Number of Samples = %d \n",npts);
	fprintf(stdout,"|   WARNING: Output will be more than 1 trace\n");
//  	fprintf(h3,"==> Large Number of Samples = %d \n",npts);
//  	fprintf(h3,"==> WARNING: Output will be more than 1 trace\n");
	}
} //endif
  
//...open output file
	if(scansw == 0 ) {
          if( (h2=fopen(ofile,"wb"))==NULL )
          {
            fprintf(stderr,"\n ABORT !! Error opening file %s",ofile);
            goto exit_1;
          }
//...dynamic allocation of seimic trace memory
        s2=calloc(npts,sizeof(float));
          if(s2==NULL)
         {
          fprintf(stderr,"ABORT:s2 can't allocate enough trace memory");
          fprintf(stderr," npts = %d",npts);
          goto exit_s1;
         }
//...re-read file and load data vector s2 with floats
	readingvalues (infile,msr,&reccount,scansw,npts,
        &year,&day,&hour,&min,&sec,
	wantStation,wantchannel,wantnetwork,wantlocation,
	wantsamprate,fsamin,s2);

//...Compute output traces length
	nptshort = 65535;
	if (npts > nptshort ) {
	nptshort = findnptshort(npts);
	if (nptshort == 0) nptshort = 65535; // if failed to find divisor <1000
	ntraces = npts / nptshort;
	int remain = npts%nptshort;
//	fprintf(h3,"==> Number of traces = %d \n",ntraces);
		if (remain != 0 ) {
		fprintf(stdout,"|   Residual data not output (modulo test) \n");
		fprintf(stdout,"|   Remainder = %d \n",remain);
//		fprintf(h3,"==> Residual data not output (modulo test) \n");
//		fprintf(h3,"==> Remainder = %d \n",remain);
		}
	}
	else {
	nptshort = npts;
	} //endif test for large trace



//...output traces
	float	*s2ptr = &s2[0];
	for (jrec=0; jrec <ntraces; jrec++) { 

//...generate a BSEGY header
	samprate = (float) wantsamprate;
	genhead(jrec,fsamin,nptshort,samprate,wantchannel,wantStation,
	year,day,hour,min,sec,&hd); 

        if(c_bsegout(jrec,s2ptr,nptshort,&hd,h2)!=0)
       {
        fprintf(stderr,"ABORT--c_bsegout");
        goto quit_it;
       }
	s2ptr = s2ptr + nptshort;
	} //next jrec

} //endif scansw

//...generate a Gnuplot file for datfile
	if (scansw == 0)
	gengnuplot(plotfile,wantStation,wantchannel,year,day,hour,min,sec,datfile);


//... alternative test for long names
	if (strlen(infile)>12) {
	puts("|   Warning: long file name, check output file name for truncations");
	fprintf(stdout,"|   InputFile: %s \n",infile);
	}
	puts("------------------------Output Files---------------------------------");
	printf(" *.lst file ====>%s \n",lstfile);
      
	if (scansw == 0) {
	printf("BSEGY *.seg file  ====>%s \n",ofile);
	printf("ASCII *.dat file ====>%s \n",datfile);
	printf("Gnuplot *.gp  file ====>%s \n",plotfile);
	}

      printf("   Processed %d records \n",reccount);
      printf("   Number of traces = %d \n",ntraces);
	free(s1);
	if(scansw == 0 ) free(s2);
        return(0);

quit_it:
        puts(" ");
        if(scansw == 0) free(s2);
exit_s1:
        free(s1);
exit_2:
        fclose(h2);
exit_1:
        fclose(h1);
exit_3:
        fclose(h3);
quick_exit:
        return(1);
     } //  end of main
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: getparm.c                                   |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: gets input parameters from command                |
   |          line or prompts user from consol.                 |
   |                                                            |
   |                                                            |
   |------------------------------------------------------------|
*/
      int  getparm(int argc,char *argv[],
         char infile[], char wantStation[], char wantchannel[], char wantnetwork[],
	 char wantlocation[], double *wantsamprate, int *scansw,
         char ofile[], char lstfile[],char datfile[],char pid[])
     {
      char              sfx1[4]="seg",
			sfx3[4]="dat",
                        sfx2[4]="lst";
      char		help[3]="-h";
      char              msg1[46],msg2[46];

//...test for command line help ?
      if(argc==2 && !strcmp(argv[1],help))
     {
//...screen header and copyright notice
      cr_labl(pid);

//...help pannel
      sprintf(msg1,"Convert MSEED to bsu BSEGY format            ");
      sprintf(msg2,"Parameters select wanted cases               ");
      hlp_labl(msg1,msg2);
      puts(" ");

//...help display
      puts(" ");
      printf("  mseed2seg  infile scansw station channel network location samprate  \n");
      puts(" ");
      puts("     infile   = input file name (4char minimum)");
      puts("     scansw =0 apply filters, extract waveform data (int)");
      puts("            =1 scan headers only, 1 line per record (int)");
      puts("            =2 scan headers only,  more detail (int)");
      puts("            =3+ scan headers only, with increasing detail (int)");
      puts("     station  = desired station name (char) ex. COR ");
      puts("     channel  = desired channel (char) ex. BHZ ");
      puts("     network  = desired network (char) ex. IU             | OR x matches all ");
      puts("     location = desired location at station (char) ex. 10 | OR x matches all");
      puts("     samprate = desired sample rate (double) ex. 40.0");
      puts(" ");
      return(1);
     }
      else
     {

//...get parameters------------------------------
//...pars off character parameters
      if(argc>1)
       strncpy(infile,argv[1],131);
       else
      {
       puts(" Enter INPUT file name ");
       scanf(" %131s",infile);
      }
//	if (namecheck(12,infile) != 0 ) return (1);

//...Test if only header scan SCANSW
	if(argc>2)
	*scansw = atoi(argv[2]);
	else
	{ 
	 puts(" Choose scan headers or also output data");
	 puts(" scansw =0 apply filters, extract waveform data (int)");
	 puts("        =1 scan headers only, 1 line per record (int)");
	 puts("        =2 scan headers only,  more detail (int)");
	 puts("        =3+ scan headers only, with increasing detail (int)");
	 scanf("%d",scansw);
	}

//...bail out if just scan
	if ( *scansw == 0 ) {


//...get desired station name 
      if(argc>3)
       strncpy(wantStation,argv[3],10);
       else
      {
       puts(" Enter desired station name (char) ex. COR");
       scanf(" %10s",wantStation);
      }

//...get desired channel    
      if(argc>4)
	strncpy(wantchannel,argv[4],10);
       else
      {
       puts(" Enter desired channel name (char) ex. BHZ");
       scanf(" %10s",wantchannel);
      }

//...get desired network   
      if(argc>5)
	strncpy(wantnetwork,argv[5],10);
       else
      {
       puts(" Enter desired network name (char) ex. IU | OR x matches all" );
       scanf(" %10s",wantnetwork);
      }


//...get desired location 
      if(argc>6)
	strncpy(wantlocation,argv[6],10);
       else
      {
       puts(" Enter desired location name (char) ex. 10 | OR x matches all ");
       scanf(" %10s",wantlocation);
      }

//...get desired samplerate Hz
      if(argc>7)
	*wantsamprate = atof(argv[7]);
       else
      {
       puts(" Enter desired sample rate Hz (double)");
       scanf(" %lf",wantsamprate);
      }

}  //end if more than scan, otherwise bail out

//...name output files
//    names(infile,pid,ofile,sfx1);
//    names(infile,pid,lstfile,sfx2);
//    names(infile,pid,datfile,sfx3);
//...output file names alternative for MiniSeed source                            
	names2(ofile,wantStation,wantchannel,sfx1);
	if(*scansw == 0) {
	names2(datfile,wantStation,wantchannel,sfx3);
	names2(lstfile,wantStation,wantchannel,sfx2);
	} else {
	names(infile,pid,lstfile,sfx2);
	}
     }  //  endif bhelp test
       return(0);
      } // end getparm

//function naming output files
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: names2                                      |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: Output file name assumes 4 char station and       |
   |          3 char channel                                    |
   |          name generated ends in suffix.                    |
   |                                                            |
   |------------------------------------------------------------|
*/
	void	names2(char filename[],char wantStation[],
	 char wantchannel[], char sfx[])
	{
	int	j;
	int 	lensta;
	for (j=0; j<13; j++) filename[j]='\0';
	lensta = strlen(wantStation);
	for (j=0;j<lensta;j++) filename[j]=wantStation[j];
	for (j=lensta;j<lensta+3;j++) filename[j]=wantchannel[j-lensta];
	filename[lensta+3] = '.';
	for (j=lensta+4;j<(lensta+7);j++) filename[j]=sfx[j-(lensta+4)];
	}

/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: outlst.c                                    |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: prints copyright header, file names, and          |
   |          echo check of input parameters in newly           |
   |          opened output listing file.                       |
   |                                                            |
   |------------------------------------------------------------|
*/
      void outlst(char infile[],
        char pid[], char ofile[], char lstfile[], int scansw,
	char wantStation[],char wantchannel[], char wantnetwork[],
	 char wantlocation[], double wantsamprate, char datfile[], FILE *h3)
     {

//...print copyright label to outlst and screen

      fcr_labl(pid,h3);

//...echo check
      fputs("\n",h3);                      //blank line
      fprintf(h3,"\n     PROGRAM: mseed2seg.c ");
      fputs("\n",h3);                      //blank line
      fputs("\n     FILES:       ",h3);
      fprintf(h3,"\n     Input   File= %s",infile);
      if(scansw == 0) {
      fprintf(h3,"\n     Output  File= %s",ofile);
      fprintf(h3,"\n     Data File  File= %s",datfile);
	}
      fprintf(h3,"\n     Listing File= %s",lstfile);
      fputs("\n",h3);                      //blank line
	if (scansw == 0 ) {
      fputs("\n     INPUT PARAMETERS: ",h3);
	fprintf(h3,"\n     Select Station = %s \n",wantStation);
	fprintf(h3,"     Select Channel = %s \n",wantchannel);
	fprintf(h3,"     Select Network = %s \n",wantnetwork);
	fprintf(h3,"     Select Location= %s \n",wantlocation);
	fprintf(h3,"     Select Sample Rate = %f \n",wantsamprate);}
	fputs("---------------------------------------------------------------------\n",h3);
      if(scansw == 0)
      fputs("\n     SELECTED RECORDS: ",h3);

      fputs("\n",h3);                      //blank line
     }  //  end outlst
//-----------------------------------------------------------------------------------------
	int scanning (char infile[], MSRecord *msr, int *reccount, int scansw)
	{
		int retcode;
	while ( (retcode = ms_readmsr (&msr, infile, reclen, NULL, NULL, 1,
			 printdata, verbose)) == MS_NOERROR )
    	{
	fprintf(stdout,"# ");	
	//msr_print (msr, verbose);
	msr_print (msr, (scansw-1)); //level of detail on scan set by scansw
	*reccount +=1;
	} //endwhile
	return (retcode);
}  //endscanning
//-----------------------------------------------------------------------------------------
	int readingdata (char infile[], MSRecord *msr, int *reccount, int scansw,
	char wantStation[], char wantchannel[], char wantnetwork[], char wantlocation[],
	double wantsamprate, float fsamin, char datfile[], int *npts, FILE *h1)
	{
	int retcode;
	int itest0 = 1;
	int itest1 = 1;
	int itest2 = 1;
	int itest3 = 1;
	int itest4 = 1;
	hptime_t START;
	double T;
	struct btime_s BTime;
	uint16_t year;
	uint16_t day;
	uint8_t hour;
	uint8_t min;
	uint8_t sec;

//... set output data
  printdata = 1;
  while ( (retcode = ms_readmsr (&msr, infile, reclen, NULL, NULL, 1,
			 printdata, verbose)) == MS_NOERROR )
    {
       /* Do something with the record here, e.g. print */
	itest0= strcmp(msr->station,wantStation);
        itest1= strcmp(msr->channel,wantchannel);
 	itest2= ((int) msr->samprate - (int) wantsamprate );
  	itest3= strcmp(msr->location,wantlocation);
	if (strncmp("x",wantlocation,1)==0 ) itest3 = 0;
	itest4= strcmp(msr->network,wantnetwork);
	if (strncmp("x",wantnetwork,1)==0 )  itest4 = 0;

       if (itest0 == 0 & itest1 == 0 & itest2 ==0 & itest3 == 0 & itest4 == 0) {
	*reccount += 1;
	fprintf(h1,"#");
       msr_print (msr, verbose);
       ms_hptime2btime(msr->starttime,&BTime);

		fprintf(h1," station %s \t",msr->station);
		fprintf(h1," channel %s \t",msr->channel);
		fprintf(h1," location %s \t",msr->location);
		fprintf(h1," samprate %.1f \t",msr->samprate);
		fprintf(h1," sequence %d \n",msr->sequence_number);
		fprintf(h1,"# network %s \t",msr->network);
		fprintf(h1," numsamples %d ",msr->numsamples);
		fprintf(h1," sampletype=%c \t",msr->sampletype);
		fprintf(h1," samplecnt %d \t",msr->samplecnt);
		fprintf(h1," encoding %d \n",msr->encoding);
		fprintf(h1,"# byteorder %d \t",msr->byteorder);
		fprintf(h1," dataquality %c \t",msr->dataquality);
                fprintf(h1," reccount=%d \n",*reccount);

		fprintf(h1,"# starttime=%d  ",msr->starttime);
		year = BTime.year;
		day = BTime.day;
		hour = BTime.hour;
		min = BTime.min;
		sec = BTime.sec;
		fprintf(h1," (year=%d ",year);
		fprintf(h1,"day=%d ",day);
		fprintf(h1," %d:",hour);
		fprintf(h1,"%d:",min);
		fprintf(h1,"%d) ",sec);
 
 
if (*reccount == 1) {
START = msr->starttime;} //microsec

		float difftime =(float) (msr->starttime - START)/1000000.;
		fprintf(h1," difftime=%f sec. \n",difftime);

    if (msr->numsamples > 0)
    {
      int  cnt, samplesize;
      void *sptr;

      if ((samplesize = ms_samplesize (msr->sampletype)) == 0)
      {
        ms_log (2, "Unrecognized sample type: '%c'\n", msr->sampletype);
      }

      for (cnt = 0; cnt < msr->numsamples; cnt++)
      {
          if (cnt < msr->numsamples)
          {
            sptr = (char *)msr->datasamples + (cnt * samplesize);

            if (msr->sampletype == 'i') {
              //ms_log (0, "%10d  ", *(int32_t *)sptr);
		T=(double)(cnt)*fsamin + difftime;
		fprintf(h1," %f  %10d \n",T,*(int32_t *)sptr);
		*npts += 1 ;
		} //end integer

            else if (msr->sampletype == 'f') {
              //ms_log (0, "%10.8g  ", *(float *)sptr);
		T=(double)(cnt)*fsamin + difftime;
		fprintf(h1," %f  %10d \n",T,*(float *)sptr);
		*npts += 1 ;
		} //end float

            else if (msr->sampletype == 'd') {
              //ms_log (0, "%10.10g  ", *(double *)sptr);
		T=(double)(cnt)*fsamin + difftime;
		fprintf(h1," %f  %10f \n",T,*(double *)sptr);
		*npts += 1 ;
		}  //end double
          }  //end if (cnt < number of samples)
       } //next cnt
     }  //end if number of samples > 0

   } //endif matches filter
 } //endwhile
return(retcode);
} //end readingdata
//-----------------------------------------------------------------------------------------



#ifndef WIN
/***************************************************************************
 * term_handler:
 * Signal handler routine.
 * From msview.c by Chad Trabant, ORFEUS/EC-Project MEREDIAN
 ***************************************************************************/
static void
term_handler (int sig)
{
  exit (0);
}
#endif

/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: genhead.c                                   |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose:  generate a 240 byte trace header                 |
   |                                                            |
   |------------------------------------------------------------|
*/
void  genhead(int jrec, float fsamin, int npts, float samprate, char wantchannel[], 
	char wantStation[], 
        uint16_t year, uint16_t day, uint8_t hour, uint8_t min, uint8_t sec,
        struct header *hd)	
{

  int  jt;
  char buf1[10];
  short moday;
//generate trace header
      (*hd).tsnl = (long) jrec + 1;
      (*hd).tsnt = (long) jrec + 1;
      (*hd).ofrn = (long) 1;
      (*hd).tnofr = (long) jrec + 1;
      (*hd).cdp = (long) 1;
      (*hd).duse = (short) 1;
      (*hd).nvs = (short) 1;
      (*hd).nhs = (short) 1;
      (*hd).tic = (short) 1;
      (*hd).ictime = (short) 0;
      (*hd).isi =  (int) (fsamin * 1000000. + 0.5);
      (*hd).length = (int) npts;
      (*hd).gaint = (short) 3;
      (*hd).gidb = (short) 0;
      (*hd).utimes = (short) 0;
      (*hd).utimeg = (short) 0;
      (*hd).charge = (short) 0;
      (*hd).smul1 = (short) -1000;
      (*hd).smul2 = (short) -1000;
      for (jt = 0; jt < 4; jt++) {
	(*hd).spname[jt] = ' ';
	(*hd).rstnam[jt] = ' ';
	}
      (*hd).spname[3] = '1';

//      sprintf (buf1, "%4.4d", jrec + 1);

      for (jt = 0; jt < 3; jt++)
	(*hd).rstnam[jt+1] = wantchannel[jt];
//	(*hd).rstnam[jt] = buf1[jt];

//      sprintf (buf1, "0001");
      for (jt = 0; jt < 4; jt++)
	(*hd).lineid[jt] = wantStation[jt];
//	(*hd).lineid[jt] = buf1[jt];
//...Because miniseed does not have a header for BH1 and BH2 azimuth, a guess
	if (wantchannel[2]=='E' || wantchannel[2]=='1') {
	(*hd).geoazi = (short) 90;}
	else {
	(*hd).geoazi = (short) 0;}

	if (wantchannel[2]=='Z') {
	 (*hd).geover = (short) 180;}  //not sure of polarity convention, up or down
	else {
	(*hd).geover = (short) 90;} //horizontal component
      
      (*hd).shtver = (short) 0; //no way to know source moment tensor from miniseed
      (*hd).shtazi = (short) 0; 
      (*hd).mst = (long) 0;
      (*hd).sday = (short) 0;
      (*hd).shour = (short) 0;
      (*hd).smin = (short) 0;
      (*hd).sseco = (short) 0;
      (*hd).tyear = (short) year;
/*    (*hd).tday = (short)  day;   // day = day of year
      program bdump.f expects this format 4 digits mmdd
      This format was used by Bison Seismographs, 
      See egg2seg.c code also impacted by this original	
      format decision and coding of bdump.f
      Function doy2moday converts from day of year to mmdd format
*/
      moday = doy2moday((int) day, (int) year);
      (*hd).tday = (short)  moday;
      (*hd).thour = (short) hour ;
      (*hd).tmin = (short)  min;
      (*hd).tsec = (short)  sec;
      (*hd).ishlo = (long) 0;
      (*hd).ishla = (long) 0;
      (*hd).ishd = (long) 0;
      (*hd).ishe = (long) 0;
      (*hd).irlo = (long) 0;
      (*hd).irla = (long) 0;
      (*hd).irel = (long) 0;
      (*hd).flc = (short) 0;
      (*hd).fhc = (short) samprate/2; //miniseed does not have filters
      }  // end genhead
//-----------------------------------------------------------------------------------------
	int readingvalues (char infile[], MSRecord *msr, int *reccount, int scansw,int npts,
        uint16_t *year, uint16_t *day, uint8_t *hour, uint8_t *min, uint8_t *sec,
	char wantStation[], char wantchannel[], char wantnetwork[], char wantlocation[],
	double wantsamprate, float fsamin, float *s2 )
	{
	int retcode;
	int itest0 = 1;
	int itest1 = 1;
	int itest2 = 1;
	int itest3 = 1;
	int itest4 = 1;
	struct btime_s BTime;
	int	index=0;
	*reccount=0;

//... set output data
  printdata = 1;
  while ( (retcode = ms_readmsr (&msr, infile, reclen, NULL, NULL, 1,
			 printdata, verbose)) == MS_NOERROR )
    {

//...bail out if too many samples
//	if (index >= npts ) break;

       /* Do something with the record here, e.g. print */
	itest0= strcmp(msr->station,wantStation);
        itest1= strcmp(msr->channel,wantchannel);
 	itest2= ((int) msr->samprate - (int) wantsamprate );
  	itest3= strcmp(msr->location,wantlocation);
	if (strncmp("x",wantlocation,1)==0 ) itest3 = 0;
	itest4= strcmp(msr->network,wantnetwork);
	if (strncmp("x",wantnetwork,1)==0 ) itest4 = 0;

       if (itest0 == 0 & itest1 == 0 & itest2 ==0 & itest3 == 0 & itest4 == 0) {
	*reccount += 1;
 
	if (*reccount == 1) {
        ms_hptime2btime(msr->starttime,&BTime);
		*year = BTime.year;
		*day = BTime.day;
		*hour = BTime.hour;
		*min = BTime.min;
		*sec = BTime.sec;
	} //microsec


    if (msr->numsamples > 0)
    {
      int  cnt, samplesize;
      void *sptr;

      if ((samplesize = ms_samplesize (msr->sampletype)) == 0)
      {
        ms_log (2, "Unrecognized sample type: '%c'\n", msr->sampletype);
      }

      for (cnt = 0; cnt < msr->numsamples; cnt++)
      {
          if (cnt < msr->numsamples)
          {
            sptr = (char *)msr->datasamples + (cnt * samplesize);

//... bail out if running out of memory
//	if (index >= npts ) break;

            if (msr->sampletype == 'i') {
		//fprintf(h1," %f  %10d \n",T,*(int32_t *)sptr);
		s2[index]=(float) (*(int32_t *)sptr);
		index += 1 ;
		} //end integer

            else if (msr->sampletype == 'f') {
              //ms_log (0, "%10.8g  ", *(float *)sptr);
		s2[index]=(float) (*(float *)sptr);
		index += 1 ;
		} //end float

            else if (msr->sampletype == 'd') {
              //ms_log (0, "%10.10g  ", *(double *)sptr);
		s2[index]=(float) (*(double *)sptr);
		index += 1 ;
		}  //end double

          }  //end if (cnt < number of samples)
       } //next cnt
     }  //end if number of samples > 0

   } //endif matches filter
 } //endwhile
return(retcode);
} //end readingvalues


//============================================================
/*
	Given day of year (doy), return an integer encoded
	as monthday.  For example moday = 0602 for June 02
*/
	short doy2moday(int doy,  int year)
	{
	int	days[12]={31,28,31,30,31,30,31,31,
				  30,31,30,31} ;
	int	j=0;
	int	sumL =0;
	int	sumU =0;
	int	month;
	int	dom;
	short	moday; // month.day

	if ((year % 4)==0  ) days[1]=29;


	while (j<12) {
	sumU = sumU + days[j];
	//...figure leap year
	if (sumU >= doy) break;
	j =j + 1;
	sumL = sumU;
	} // endwhile next j

//...found
	dom = doy - sumL;
	month = j+1;
	moday = month*100+dom;
	return(moday);
	} //end doy2moday
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: gengnuplot                                  |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: Generates a gnuplot script to plot *.dat outfile  |
   |                                                            |
   |------------------------------------------------------------|
*/
	void	gengnuplot(char plotfile[], char wantStation[], char wantchannel[],
	uint16_t year, uint16_t day, uint8_t hour, uint8_t min, uint8_t sec,
	char datfile[])

	{
	FILE	*h4;
	char	sfx[4]="gp";
	char	sfxps[4]="ps";
	char	station[4];
	char	channel[4];
	char	title[80];
	char	legend[40];
	char	postscriptfile[13];
	short	moday;

//...name output Gnuplot file
	names2(plotfile,wantStation,wantchannel,sfx);
	plotfile[12]='\0';
	names2(postscriptfile,wantStation,wantchannel,sfxps);
	postscriptfile[12]='\0';

//...set up start time and date for record first sample
//      moday = mmdd  ex. 1220 = December 20
  	moday = doy2moday((int) day, (int) year);
	int month = moday/100;
	int Day = moday-month*100;
	
//...set up title
	sprintf(title,"Station: %s    Channel: %s (%2.2d:%2.2d:%2.2d) %2.2d/%2.2d/%2.2d",
	wantStation,wantchannel,hour,min,sec,month,Day,year);

//...set up legend for channel orientation
	switch(wantchannel[2]) {
	case 'Z' :
	sprintf(legend,"Vertical Motion");
	break;
	case 'E' :
	sprintf(legend,"East Motion");
	break;
	case 'N' :
	sprintf(legend,"North Motion");
	break;
	case '1' :
	sprintf(legend,"Horizontal 1 Motion");
	break;
	case '2' :
	sprintf(legend,"Horizontal 2 Motion");
	break;
	default :
	sprintf(legend," ");
	}  //endswitch

//...open output plotfile file
        if( (h4=fopen(plotfile,"w"))==NULL )
        {
          fprintf(stderr,"\n ABORT !! Error opening file %s",plotfile);
          abort();
        }
	fprintf(h4,"#Generated by mseed2seg.c \n");
	fprintf(h4,"# Default is live plot \n");
	fprintf(h4,"# For fixed image file remove # on TYPE=image \n");
	fprintf(h4,"# and rerun:  gnuplot %s \n",plotfile);
#ifdef WIN
  	fprintf(h4,"TYPE=\"live\" \n"); 
  	fprintf(h4,"#TYPE=\"image\" \n"); 
  	fprintf(h4,"if (TYPE ne \"live\") { \n"); 
  	fprintf(h4,"set terminal jpeg large size 800,600 enhanced \n"); 
  	fprintf(h4,"set output \"%s\" \n",postscriptfile); 
#else
  	fprintf(h4,"TYPE=\"live\" \n"); 
  	fprintf(h4,"#TYPE=\"image\" \n"); 
  	fprintf(h4,"if (TYPE ne \"live\") { \n"); 
  	fprintf(h4,"set terminal postscript size 8.0,6.0 in color \\\n"); 
	fprintf(h4,"portrait enhanced font \'Helvetica,10\' \n");
  	fprintf(h4,"set output \"%s\" \n",postscriptfile); 
#endif
  	fprintf(h4," } else { \n"); 
	fprintf(h4,"TRM=GPVAL_TERM                   \n");
	fprintf(h4,"if (TRM eq \"unknown\" ) {set terminal x11 persist} \n");
	fprintf(h4,"if (TRM eq \"qt\") { set terminal wxt size 800,600 persist  } \n");
	fprintf(h4,"if (TRM eq \"wxt\") { set terminal wxt size 800,600 persist }   \n");
	fprintf(h4,"if (TRM eq \"x11\") { set terminal x11 persist } \n");
  	fprintf(h4," }  \n"); 
	fprintf(h4,"                                 \n");
	fprintf(h4,"set grid \n");
	fprintf(h4,"set title \"%s\" \n",title); 
	fprintf(h4,"set xlabel \"Time (sec)\" \n");
	fprintf(h4,"set ylabel \"Amplitude\" \n");
	fprintf(h4,"p '%s' u 1:2 w l lc rgb \"blue\" t \"%s\" \n",datfile,legend);
//	fprintf(h4,"set terminal postscript portrait color size 7,4 \n");
//	fprintf(h4,"set output \"%s\" \n",postscriptfile);
//	fprintf(h4,"replot \n");
	fclose(h4);
	} //end gengnuplot
/* |------------------------------------------------------------|
   | Basic Seismic Utilities (BSU)                              | 
   | LANGUAGE: C                                                |
   | AUTHOR: P. MICHAELS <paulmichaels@boisestate.edu>          |
   | FUNCTION NAME: findnptshort                                |
   | Copyright (C) 2021 P. Michaels  (See GNU License)          |
   |------------------------------------------------------------|
   |                                                            |
   | purpose: find a short value to break up trace into sub     |
   |          traces.                                           |
   |------------------------------------------------------------|
*/
	int  findnptshort(int npts)
	{
	int	nmax = 65535;
	int	j=1;
	int	j2;
	int	nptshort;
	int	remain,
		itest;
	int	jsave;
	nptshort = npts;
	while (nptshort > nmax) {
	j = j + 1;
	nptshort = npts/j;
		if (j>1000) {
		return(0);
		}
	} //endwhile

//...look for a smaller residual
	jsave = j;
	itest = npts%nptshort;
	for (j2=1; j2<10; j2++) {
	remain = npts%(npts/(j+j2));
		if(remain < itest ) {
		itest=remain;
		jsave = j+j2;
		} //endif
	} //next j2	
	nptshort = npts / jsave;
	return(nptshort);
	} //end findnptshort
